// See LICENSE.SiFive for license details.
// See LICENSE.Berkeley for license details.
//
// Author: Shuwen Deng <shuwen.deng@yale.edu>
// Date:   08.27.2019
//
// This file implements RF TLB for secure TLB project
//
// This code is released under same license as RocketChip code,
// license details can be found in LICENSE.SiFive file.
//

package freechips.rocketchip.rocket

import Chisel._
import Chisel.ImplicitConversions._

import freechips.rocketchip.config.{Field, Parameters}
import freechips.rocketchip.subsystem.CacheBlockBytes
import freechips.rocketchip.diplomacy.RegionType
import freechips.rocketchip.tile.{XLen, CoreModule, CoreBundle}
import freechips.rocketchip.tilelink._
import freechips.rocketchip.util._
import freechips.rocketchip.util.property._
import chisel3.internal.sourceinfo.SourceInfo

//case object PgLevels extends Field[Int](2)
//case object ASIdBits extends Field[Int](8)

class RERF_SFenceReq(implicit p: Parameters) extends CoreBundle()(p) {
  val rs1 = Bool()
  val rs2 = Bool()
  val addr = UInt(width = vaddrBits)
  val asid = UInt(width = asIdBits max 1) // TODO zero-width
}

class RERF_TLBReq(lgMaxSize: Int)(implicit p: Parameters) extends CoreBundle()(p) {
  val vaddr = UInt(width = vaddrBitsExtended)
  val passthrough = Bool()
  val sfence = Valid(new RERF_SFenceReq)
  val size = UInt(width = log2Ceil(lgMaxSize + 1))
  val cmd  = Bits(width = M_SZ)

  override def cloneType = new RERF_TLBReq(lgMaxSize).asInstanceOf[this.type]
}

class RERF_TLBExceptions extends Bundle {
  val ld = Bool()
  val st = Bool()
  val inst = Bool()
}

class RERF_TLBResp(implicit p: Parameters) extends CoreBundle()(p) {
  // lookup responses
  val miss = Bool()
  val paddr = UInt(width = paddrBits)
  val pf = new RERF_TLBExceptions
  val ae = new RERF_TLBExceptions
  val ma = new RERF_TLBExceptions
  val cacheable = Bool()
  val prefetchable = Bool()
  //SD: for evict and time
  val not_miss_and_R_sec = Bool()
}

class RERF_TLB(instruction: Boolean, lgMaxSize: Int, nEntries: Int)(implicit edge: TLEdgeOut, p: Parameters) extends CoreModule()(p) with HasL1HellaCacheParameters{
  // SD
  val secBits:Int = 1
  val bufferSize:Int = 1
  val io = new Bundle {
    val req = Decoupled(new RERF_TLBReq(lgMaxSize)).flip
    val resp = new RERF_TLBResp().asOutput
    val ptw = new TLBPTWIO
    // SD: RF cache
    val random_refill_start = Bool()
    val random_fill_inst_valid = Bool().asInput
    val random_fill_region = Bool().asInput
    // SD: edit for no fill
    val no_fill = Bool().asInput
    val no_fill_end_tlb = Bool().asInput
    val whole_random_fill_region = Bool().asInput
    val dcache_end = Bool().asInput
    val random_fill_counter = UInt(width=2).asInput
  }

  class RERF_Entry extends Bundle {
    val ppn = UInt(width = ppnBits)
    // SD
    val tag = UInt(width = asIdBits + secBits + vpnBits - log2Ceil(nSetsTLB))
    //val index = UInt(width = nSetsTLB)
    val level = UInt(width = log2Ceil(pgLevels))
    val u = Bool()
    val g = Bool()
    val ae = Bool()
    val sw = Bool()
    val sx = Bool()
    val sr = Bool()
    val pw = Bool()
    val px = Bool()
    val pr = Bool()
    val pal = Bool() // AMO logical
    val paa = Bool() // AMO arithmetic
    val eff = Bool() // get/put effects
    val c = Bool()
  }
  // SD
  val totalEntries = nWaysTLB + 1
  val normalEntries = nWaysTLB
  val specialEntry = nWaysTLB //+ bufferSize
  val aeEntry = specialEntry /*- bufferSize*/ /*problem*/- (1 << log2Floor(nWaysTLB + bufferSize))
  val valid = Reg(init = Vec(Seq.fill(nSetsTLB)(UInt(0, totalEntries+bufferSize))))
    //Reg(Vec(Seq.fill(nSetsTLB)(init = UInt(0, totalEntries))))
  // SD: for no fill buffer
  val reg_entries = Reg(Vec(nSetsTLB, Vec(totalEntries+bufferSize/*buffer*/, UInt(width = new RERF_Entry().getWidth))))
  val index_entries = Reg(Vec(nSetsTLB, Vec(totalEntries+bufferSize, new RERF_Entry())))
  for (i <- 0 until nSetsTLB) {
    for (j <- 0 until totalEntries+bufferSize){
      index_entries(i)(j) := reg_entries(i)(j).asTypeOf(new RERF_Entry)
    }
  }

  val s_ready :: s_request :: s_wait :: s_wait_invalidate :: Nil = Enum(UInt(), 4)
  val state = Reg(init=s_ready)
  // SD
  val r_refill_tag = Reg(UInt(width = asIdBits + secBits + vpnBits - log2Ceil(nSetsTLB)))
  val r_refill_waddr = Reg(UInt(width = log2Ceil(normalEntries+bufferSize)))
  val r_refill_index = Reg(UInt(width = log2Ceil(nSetsTLB) max 1))
  val r_refill_cmd = Reg(UInt(width = M_SZ))
  val r_req = Reg(new RERF_TLBReq(lgMaxSize))

  // SD: store the refill index and waddr for no fill buffer
  val r_nofill_index = Reg(UInt(width = log2Ceil(nSetsTLB) max 1))


  val priv = if (instruction) io.ptw.status.prv else io.ptw.status.dprv
  val priv_s = priv(0)
  val priv_uses_vm = priv <= PRV.S
  val vm_enabled = Bool(usingVM) && io.ptw.ptbr.mode(io.ptw.ptbr.mode.getWidth-1) && priv_uses_vm && !io.req.bits.passthrough
  val tag = io.req.bits.vaddr(vaddrBits-1, pgIdxBits + log2Ceil(nSetsTLB))
  // SD: for evict and time
  val cmd = io.req.bits.cmd
  def abs_val(x1:Int, x2:Int) = if (x1>=x2) x1-x2 else x2-x1
  val index = Mux (/*(!vm_enabled)||*/nSetsTLB==1, 0.U, io.req.bits.vaddr(pgIdxBits + abs_val(log2Ceil(nSetsTLB), 1), pgIdxBits))
  //val index = Mux (/*(!vm_enabled)||*/nSetsTLB==1, 0.U, Mux(nSetsTLB==2, io.req.bits.vaddr(pgIdxBits), io.req.bits.vaddr(pgIdxBits + abs_val(log2Ceil(nSetsTLB), 1), pgIdxBits)))

  // refill waddr
  val waddr_out = RegInit(UInt(0, log2Ceil(normalEntries+bufferSize)))


  // share a single physical memory attribute checker (unshare if critical path)
  val refill_ppn = io.ptw.resp.bits.pte.ppn(ppnBits-1, 0)
  val do_refill = Bool(usingVM) && io.ptw.resp.valid
  val invalidate_refill = state.isOneOf(s_request /* don't care */, s_wait_invalidate)
  // SD: edit for buffer
  val mpu_ppn = Mux(do_refill, refill_ppn,
                Mux(vm_enabled, index_entries(/*r_refill_index*/index)/*(nWaysTLB+bufferSize)
                  */.last.ppn, io.req.bits.vaddr >> pgIdxBits))
  val mpu_physaddr = Cat(mpu_ppn, io.req.bits.vaddr(pgIdxBits-1, 0))
  val pmp = Module(new PMPChecker(lgMaxSize))
  pmp.io.addr := mpu_physaddr
  pmp.io.size := io.req.bits.size
  pmp.io.pmp := (io.ptw.pmp: Seq[PMP])
  pmp.io.prv := Mux(Bool(usingVM) && (do_refill || io.req.bits.passthrough /* PTW */), PRV.S, priv)
  // SD: for TLB random address access
  val legal_address = (edge.manager.findSafe(mpu_physaddr).reduce(_||_)) //|| (io.random_fill_region)
  def fastCheck(member: TLManagerParameters => Boolean) =
    legal_address && edge.manager.fastProperty(mpu_physaddr, member, (b:Boolean) => Bool(b))
  val cacheable = fastCheck(_.supportsAcquireT) && (instruction || !usingDataScratchpad)
  val homogeneous = TLBPageLookup(edge.manager.managers, xLen, p(CacheBlockBytes), BigInt(1) << pgIdxBits)(mpu_physaddr).homogeneous
  val prot_r = fastCheck(_.supportsGet) && pmp.io.r
  val prot_w = fastCheck(_.supportsPutFull) && pmp.io.w
  val prot_al = fastCheck(_.supportsLogical) || (cacheable && usingAtomicsInCache)
  val prot_aa = fastCheck(_.supportsArithmetic) || (cacheable && usingAtomicsInCache)
  val prot_x = fastCheck(_.executable) && pmp.io.x
  val prot_eff = fastCheck(Seq(RegionType.PUT_EFFECTS, RegionType.GET_EFFECTS) contains _.regionType)

  // SD
  val vpn = io.req.bits.vaddr(vaddrBits-1, pgIdxBits)
  val lookup_tag = Mux(asIdBits.U===0.U, tag, Cat(io.ptw.ptbr.asid(asIdBits-1, 0), /*io.ptw.ptbr.sec(secBits-1,0)*//*io.random_fill_inst_valid*/io.whole_random_fill_region.asUInt /*(io.random_fill_inst_valid||io.random_fill_region).asUInt*/, tag)) //vpn)
  // SD: edit for buffer
  val hitsVec = (0 until (totalEntries+bufferSize)).map { i => if (!usingVM) false.B else vm_enabled && {
    var tagMatch = valid(index)(i)
    for (j <- 0 until pgLevels) {
      val base = vpnBits - (j + 1) * pgLevelBits
      //tagMatch = tagMatch && (index_entries(index)(i).level < j
      if (j+1 == pgLevels) {
        tagMatch = tagMatch && (index_entries(index)(i).level < j || index_entries(index)(i).tag(base + pgLevelBits - 1 - log2Ceil(nSetsTLB), base) === vpn(base + pgLevelBits - 1, base + log2Ceil(nSetsTLB)))
      } else {
        tagMatch = tagMatch && ((index_entries(index))(i).level < j || index_entries(index)(i).tag(base + pgLevelBits - 1 - log2Ceil(nSetsTLB), base - log2Ceil(nSetsTLB)) === vpn(base + pgLevelBits - 1, base))
      }
    }
    val sec_width=1
    val D_id_in = io.ptw.ptbr.asid(asIdBits-1, 0)/*tag(asIdBits + vpnBits - log2Ceil(nSetsTLB)-1, vpnBits - log2Ceil(nSetsTLB)) + 0.U*/
    val R_id_in = index_entries(index)(i).tag(asIdBits + vpnBits - log2Ceil(nSetsTLB)-1+sec_width, vpnBits - log2Ceil(nSetsTLB)+sec_width) + 0.U
    val idMatch = Mux(index_entries(index)(i).tag(vpnBits-1/*+sec_width*/- log2Ceil(nSetsTLB))===1.U, true.B, D_id_in === R_id_in)
    tagMatch = tagMatch && idMatch
    tagMatch
  }} :+ !vm_enabled
  val hits = hitsVec.asUInt
  val level = Mux1H(hitsVec.init, index_entries(index).map(_.level))
  val partialPPN = Mux1H(hitsVec.init, index_entries(index).map(_.ppn))
  val ppn = if (!usingVM) vpn else {
    var ppn = Mux(vm_enabled, partialPPN, vpn)(pgLevelBits*pgLevels - 1, pgLevelBits*(pgLevels - 1))
    for (i <- 1 until pgLevels)
      ppn = Cat(ppn, (Mux(level < i, vpn, 0.U) | partialPPN)(vpnBits - i*pgLevelBits - 1, vpnBits - (i + 1)*pgLevelBits))
    ppn
  }
  //SD: for random fill of evict and time
  val not_miss_and_R_sec_ori = RegInit(false.B)
    //(r_refill_cmd===M_XRD_NORM||r_refill_cmd===M_XWR_NORM) && ((index_entries(r_refill_index)(r_refill_waddr).tag(secBits + vpnBits - log2Ceil(nSetsTLB)-1)=/=1.U) && valid(r_refill_index)(r_refill_waddr)===1.U)
  val not_miss_and_R_sec_flag = RegNext(not_miss_and_R_sec_ori)
  val not_miss_and_R_sec = not_miss_and_R_sec_ori && (!not_miss_and_R_sec_flag)
  //RegInit(false.B)

  // permission bit arrays
  when (do_refill && !invalidate_refill) {
    // SD: edit for bufferSize

    val waddr = Mux(
      io.ptw.resp.bits.ae, aeEntry.U, Mux(!io.ptw.resp.bits.homogeneous, specialEntry.U, /*r_refill_waddr*/
      //SD: for evict and time
      Mux(io.no_fill && (!not_miss_and_R_sec_ori)/*||(
          (r_refill_cmd===M_XRD_NORM||r_refill_cmd===M_XWR_NORM) && ((index_entries(r_refill_index)(r_refill_waddr).tag(secBits + vpnBits - log2Ceil(nSetsTLB)-1)===1.U) && valid(r_refill_index)(r_refill_waddr)===1.U))*/,
      (totalEntries).U, r_refill_waddr
         )
                                          )
                  )
    waddr_out := waddr + 0.U

    val pte = io.ptw.resp.bits.pte
    val newEntry = Wire(new RERF_Entry)
    newEntry.ppn := pte.ppn
    newEntry.tag := r_refill_tag //Mux(io.no_fill && (!not_miss_and_R_sec_ori), r_refill_tag &(~((UInt(1, lookup_tag.getWidth-tag.getWidth))<<(tag.getWidth))), r_refill_tag)
      //r_refill_tag 
    //Mux(io.random_fill_region, random_refill_tag, r_refill_tag)
    newEntry.level := io.ptw.resp.bits.level
    newEntry.c := cacheable
    newEntry.u := pte.u
    newEntry.g := pte.g
    newEntry.ae := io.ptw.resp.bits.ae
    newEntry.sr := pte.sr()
    newEntry.sw := pte.sw()
    newEntry.sx := pte.sx()
    newEntry.pr := prot_r && !io.ptw.resp.bits.ae
    newEntry.pw := prot_w && !io.ptw.resp.bits.ae
    newEntry.px := prot_x && !io.ptw.resp.bits.ae
    newEntry.pal := prot_al
    newEntry.paa := prot_aa
    newEntry.eff := prot_eff

    when(io.no_fill && (!not_miss_and_R_sec_ori)){
      r_nofill_index := r_refill_index
    }

    
    valid(r_refill_index) := valid(r_refill_index) | UIntToOH(waddr)
    reg_entries(r_refill_index)(waddr) := newEntry.asUInt
  }.elsewhen(io.dcache_end){
    not_miss_and_R_sec_ori := false.B
  }

  when(io.no_fill_end_tlb){
    valid(r_nofill_index) := valid(r_nofill_index) & (~UIntToOH((totalEntries).U))
  }

  val plru = new nSetsPseudoLRU(normalEntries, nSetsTLB)
  //val plru = new PseudoLRU(normalEntries)
  val repl_waddr = Mux(!valid(index)(normalEntries-1, 0).andR, PriorityEncoder(~valid(index)(normalEntries-1, 0)), plru.replace(index)/*(nWaysTLB-1).U*/)

  val ptw_ae_array = index_entries(index)(aeEntry).ae << aeEntry
  val priv_rw_ok = Mux(!priv_s || io.ptw.status.sum, index_entries(index).map(_.u).asUInt, 0.U) | Mux(priv_s, ~index_entries(index).map(_.u).asUInt, 0.U)
  val priv_x_ok = Mux(priv_s, ~index_entries(index).map(_.u).asUInt, index_entries(index).map(_.u).asUInt)
  val r_array = Cat(true.B, priv_rw_ok & (index_entries(index).map(_.sr).asUInt | Mux(io.ptw.status.mxr, index_entries(index).map(_.sx).asUInt, UInt(0))))
  val w_array = Cat(true.B, priv_rw_ok & index_entries(index).map(_.sw).asUInt)
  val x_array = Cat(true.B, priv_x_ok & index_entries(index).map(_.sx).asUInt)
  val pr_array = Cat(Fill(2, prot_r), index_entries(index).init.map(_.pr).asUInt)
  val pw_array = Cat(Fill(2, prot_w), index_entries(index).init.map(_.pw).asUInt)
  val px_array = Cat(Fill(2, prot_x), index_entries(index).init.map(_.px).asUInt)
  val paa_array = Cat(Fill(2, prot_aa), index_entries(index).init.map(_.paa).asUInt)
  val pal_array = Cat(Fill(2, prot_al), index_entries(index).init.map(_.pal).asUInt)
  val eff_array = Cat(Fill(2, prot_eff), index_entries(index).init.map(_.eff).asUInt)
  val c_array = Cat(Fill(2, cacheable), index_entries(index).init.map(_.c).asUInt)
  val prefetchable_array = Cat(cacheable && homogeneous, false.B, index_entries(index).init.map(_.c).asUInt)

  val misaligned = (io.req.bits.vaddr & (UIntToOH(io.req.bits.size) - 1)).orR
  val bad_va = vm_enabled &&
    (if (vpnBits == vpnBitsExtended) Bool(false)
     else (io.req.bits.vaddr.asSInt < 0.S) =/= (vpn.asSInt < 0.S))

  val lrscAllowed = Mux(Bool(usingDataScratchpad || usingAtomicsOnlyForIO), 0.U, c_array)
  val ae_array =
    Mux(misaligned, eff_array, 0.U) |
    Mux(Bool(usingAtomics) && io.req.bits.cmd.isOneOf(M_XLR, M_XSC), ~lrscAllowed, 0.U)
  val ae_ld_array = Mux(isRead(io.req.bits.cmd), ae_array | ~pr_array, 0.U)
  val ae_st_array =
    Mux(isWrite(io.req.bits.cmd), ae_array | ~pw_array, 0.U) |
    Mux(Bool(usingAtomics) && isAMOLogical(io.req.bits.cmd), ~pal_array, 0.U) |
    Mux(Bool(usingAtomics) && isAMOArithmetic(io.req.bits.cmd), ~paa_array, 0.U)
  val ma_ld_array = Mux(misaligned && isRead(io.req.bits.cmd), ~eff_array, 0.U)
  val ma_st_array = Mux(misaligned && isWrite(io.req.bits.cmd), ~eff_array, 0.U)
  val pf_ld_array = Mux(isRead(io.req.bits.cmd), ~(r_array | ptw_ae_array), 0.U)
  val pf_st_array = Mux(isWrite(io.req.bits.cmd), ~(w_array | ptw_ae_array), 0.U)
  val pf_inst_array = ~(x_array | ptw_ae_array)

  // SD: edit for buffer
  val tlb_hit = hits(totalEntries-1+bufferSize, 0).orR ///|| hits(totalEntries+bufferSize)
  val tlb_miss = vm_enabled && !bad_va && !tlb_hit && !io.req.bits.sfence.valid
  // SD: recording the start of an TLB related instruction
  val vpn_when_tlb_miss = RegInit(UInt(0, vaddrBits-pgIdxBits))
  val to_refill = io.req.fire() && tlb_miss
  val to_refill_delay = RegNext(to_refill)
  when (to_refill_delay) {
    not_miss_and_R_sec_ori := (r_refill_cmd===M_XRD_NORM||r_refill_cmd===M_XWR_NORM) && (((index_entries(r_refill_index)(r_refill_waddr).tag(secBits + vpnBits - log2Ceil(nSetsTLB)-1)=/=1.U) && valid(r_refill_index)(r_refill_waddr)===1.U)||(valid(r_refill_index)(r_refill_waddr)===0.U)) && (io.random_fill_counter===2.U||io.random_fill_counter===1.U)
  }
  when (to_refill && tlb_miss){
    vpn_when_tlb_miss := vpn
  }

  when (io.req.valid && !tlb_miss && !hits(specialEntry)) {
    plru.access(OHToUInt(hits(normalEntries-1, 0)), index)
  }

  // Superpages create the possibility that two entries in the TLB may match.
  // This corresponds to a software bug, but we can't return complete garbage;
  // we must return either the old translation or the new translation.  This
  // isn't compatible with the Mux1H approach.  So, flush the TLB and report
  // a miss on duplicate entries.
  //SD: buffer 
  val multipleHits = PopCountAtLeast(hits(totalEntries-1+bufferSize, 0), 2)

  io.random_refill_start := /*io.ptw.ptbr.sec(0).toBool*/io.random_fill_inst_valid && (to_refill && tlb_miss && (vpn_when_tlb_miss =/= vpn))
  io.req.ready := state === s_ready
  io.resp.pf.ld := (bad_va && isRead(io.req.bits.cmd)) || (pf_ld_array & hits).orR
  io.resp.pf.st := (bad_va && isWrite(io.req.bits.cmd)) || (pf_st_array & hits).orR
  io.resp.pf.inst := bad_va || (pf_inst_array & hits).orR
  io.resp.ae.ld := (ae_ld_array & hits).orR
  io.resp.ae.st := (ae_st_array & hits).orR
  io.resp.ae.inst := (~px_array & hits).orR
  io.resp.ma.ld := (ma_ld_array & hits).orR
  io.resp.ma.st := (ma_st_array & hits).orR
  io.resp.ma.inst := false // this is up to the pipeline to figure out
  io.resp.cacheable := (c_array & hits).orR
  io.resp.prefetchable := (prefetchable_array & hits).orR && edge.manager.managers.forall(m => !m.supportsAcquireB || m.supportsHint)
  //SD: for random fill
  io.resp.miss := (do_refill || tlb_miss || multipleHits) //|| (io.resp_time && io.random_fill_region)  
  io.resp.paddr := Cat(ppn, io.req.bits.vaddr(pgIdxBits-1, 0))

  //SD: for evict and time
  io.resp.not_miss_and_R_sec := not_miss_and_R_sec
  

  io.ptw.req.valid := state === s_request
  io.ptw.req.bits <> io.ptw.status
  io.ptw.req.bits.addr := Mux(nSetsTLB==1, r_refill_tag, Cat (r_refill_tag, r_refill_index))
  //Mux(io.random_fill_region, Mux(nSetsTLB==1, random_refill_tag, Cat (random_refill_tag, r_refill_index)), Mux(nSetsTLB==1, r_refill_tag, Cat (r_refill_tag, r_refill_index)))

  if (usingVM) {
    val sfence = io.req.valid && io.req.bits.sfence.valid
    when (to_refill) {
      state := s_request
      r_refill_tag := lookup_tag
      r_refill_waddr := repl_waddr
      // SD
      r_refill_index := index
      // SD: for evict and time
      r_refill_cmd := cmd
      r_req := io.req.bits
    }
    when (state === s_request) {
      when (sfence) { state := s_ready }
      when (io.ptw.req.ready) { state := Mux(sfence, s_wait_invalidate, s_wait) }
    }
    when (state === s_wait && sfence) {
      state := s_wait_invalidate
    }
    when (io.ptw.resp.valid) {
      state := s_ready
    }

    /*when (sfence) {
      assert((io.req.bits.sfence.bits.addr >> pgIdxBits) === vpn)
      val index_for_sfence = Mux (nSetsTLB==1, 0.U, io.req.bits.sfence.bits.addr(pgIdxBits + abs_val(log2Ceil(nSetsTLB), 1), pgIdxBits))
      // SD: edit for buffer
      when (!(io.random_fill_counter===0.U && ((hits(totalEntries-1+bufferSize, 0) & (1.U<<totalEntries-1+bufferSize))===0.U))){//(~UInt(1, totalEntries+bufferSize))
            valid(index_for_sfence) := Mux(io.req.bits.sfence.bits.rs1, valid(index_for_sfence) & ~hits(totalEntries-1+bufferSize, 0),
               Mux(io.req.bits.sfence.bits.rs2, valid(index_for_sfence) & index_entries(index_for_sfence).map(_.g).asUInt, 0))
        }
    }*/
    val index_for_sfence = Mux (nSetsTLB==1,0.U, io.req.bits.sfence.bits.addr(pgIdxBits + abs_val(log2Ceil(nSetsTLB), 1), pgIdxBits))
    val enter_sfence = !(io.random_fill_counter===0.U && ((hits(totalEntries-1+bufferSize, 0) & (1.U<<totalEntries-1+bufferSize))===0.U))
    for (j <- 0 until (nSetsTLB)){
      when(enter_sfence && sfence && io.req.bits.sfence.bits.rs1 && index_for_sfence===j.U) {
        valid(j) := valid(j) & ~hits(totalEntries-1, 0)
      } .elsewhen(enter_sfence && sfence && io.req.bits.sfence.bits.rs2){
        valid(j) := valid(j) & index_entries(j).map(_.g).asUInt
      } .elsewhen(enter_sfence && sfence && (!io.req.bits.sfence.bits.rs1) && (!io.req.bits.sfence.bits.rs2)){
        valid(j) := 0 
      }
    }
    when (multipleHits) {
      //valid(index) := 0
      for (j <- 0 until (nSetsTLB)){
        valid(j) := 0
      }
    }

    ccover(io.ptw.req.fire(), "MISS", "TLB miss")
    ccover(io.ptw.req.valid && !io.ptw.req.ready, "PTW_STALL", "TLB miss, but PTW busy")
    ccover(state === s_wait_invalidate, "SFENCE_DURING_REFILL", "flush TLB during TLB refill")
    ccover(sfence && !io.req.bits.sfence.bits.rs1 && !io.req.bits.sfence.bits.rs2, "SFENCE_ALL", "flush TLB")
    ccover(sfence && !io.req.bits.sfence.bits.rs1 && io.req.bits.sfence.bits.rs2, "SFENCE_ASID", "flush TLB ASID")
    ccover(sfence && io.req.bits.sfence.bits.rs1 && !io.req.bits.sfence.bits.rs2, "SFENCE_LINE", "flush TLB line")
    ccover(sfence && io.req.bits.sfence.bits.rs1 && io.req.bits.sfence.bits.rs2, "SFENCE_LINE_ASID", "flush TLB line/ASID")
    ccover(multipleHits, "MULTIPLE_HITS", "Two matching translations in TLB")
  }

  def ccover(cond: Bool, label: String, desc: String)(implicit sourceInfo: SourceInfo) =
    cover(cond, s"${if (instruction) "I" else "D"}TLB_$label", "MemorySystem;;" + desc)
}

